#include <engine/3d/collision.h>
#include <engine/asset.h>
#include <engine/math.h>

/*
 * dgx_col_cylinder_gfx:
 *	Determines if a point is within a cylinder.
 */
dgx_col_cylinder_gfx(p_gfx, p_point, radius)
struct CylinderGFX *p_gfx;
SVECTOR *p_point;
int radius;
{
	int i;

	/* FIXME: would it be faster to just take the full 3d difference then constrain that into 2d? */
	int r;
	VECTOR vec = {p_point->vx, p_point->vy, p_point->vz};
	VECTOR c_origin = {p_gfx->cyl.p[0].vx, p_gfx->cyl.p[0].vy, p_gfx->cyl.p[0].vz};

	/*
	 * Constrain the point into the plane that contains both circles of the cylinder (when
	 * viewed orthogonally through a_hat). This can be summarized as:
	 *	v(b_hat) + v(c_hat)
	 */
	{
		VECTOR tmp;
		setVector(&tmp, (vec.vx*p_gfx->endc[0].unit[0].vx)>>FIXED_PRECISION,
			(vec.vy*p_gfx->endc[0].unit[0].vy)>>FIXED_PRECISION,
			(vec.vz*p_gfx->endc[0].unit[0].vz)>>FIXED_PRECISION);

		setVector(&vec, ((vec.vx*p_gfx->endc[0].unit[1].vx)>>FIXED_PRECISION) + tmp.vx,
			((vec.vy*p_gfx->endc[0].unit[1].vy)>>FIXED_PRECISION) + tmp.vy,
			((vec.vz*p_gfx->endc[0].unit[1].vz)>>FIXED_PRECISION) + tmp.vz);

		/* now constrain the origin of one of the circles */
		setVector(&tmp, (c_origin.vx*p_gfx->endc[0].unit[0].vx)>>FIXED_PRECISION,
			(c_origin.vy*p_gfx->endc[0].unit[0].vy)>>FIXED_PRECISION,
			(c_origin.vz*p_gfx->endc[0].unit[0].vz)>>FIXED_PRECISION);

		setVector(&c_origin, ((c_origin.vx*p_gfx->endc[0].unit[1].vx)>>FIXED_PRECISION) + tmp.vx,
			((c_origin.vy*p_gfx->endc[0].unit[1].vy)>>FIXED_PRECISION) + tmp.vy,
			((c_origin.vz*p_gfx->endc[0].unit[1].vz)>>FIXED_PRECISION) + tmp.vz);
	}

	/* store the difference into vec */
	setVector(&vec, vec.vx-c_origin.vx, vec.vy-c_origin.vy, vec.vz-c_origin.vz);

	r = psqrt( ((vec.vx*vec.vx)>>FIXED_PRECISION) +
		((vec.vy*vec.vy)>>FIXED_PRECISION) +
		((vec.vz*vec.vz)>>FIXED_PRECISION) );

	FntPrint("\nr == %d\n", r);
	FntPrint("radius == %d\n", radius);
	FntPrint("radius+scale == %d\n\n", radius+p_gfx->endc[0].ent.scale);

	if(r <= radius+p_gfx->endc[0].ent.scale)
	{
		int r1, r2;
		VECTOR c_origin2 = {p_gfx->cyl.p[1].vx, p_gfx->cyl.p[1].vy, p_gfx->cyl.p[1].vz};
		setVector(&vec, p_point->vx, p_point->vy, p_point->vz);
		setVector(&c_origin, p_gfx->cyl.p[0].vx, p_gfx->cyl.p[0].vy, p_gfx->cyl.p[0].vz);

		setVector(&vec, (vec.vx*p_gfx->unit.vx)>>FIXED_PRECISION,
			(vec.vy*p_gfx->unit.vy)>>FIXED_PRECISION,
			(vec.vz*p_gfx->unit.vz)>>FIXED_PRECISION);

		setVector(&c_origin, (c_origin.vx*p_gfx->unit.vx)>>FIXED_PRECISION,
			(c_origin.vy*p_gfx->unit.vy)>>FIXED_PRECISION,
			(c_origin.vz*p_gfx->unit.vz)>>FIXED_PRECISION);

		setVector(&c_origin2, (c_origin2.vx*p_gfx->unit.vx)>>FIXED_PRECISION,
			(c_origin2.vy*p_gfx->unit.vy)>>FIXED_PRECISION,
			(c_origin2.vz*p_gfx->unit.vz)>>FIXED_PRECISION);

		r = psqrt( ((vec.vx*vec.vx)>>FIXED_PRECISION) +
			((vec.vy*vec.vy)>>FIXED_PRECISION) +
			((vec.vz*vec.vz)>>FIXED_PRECISION) );

		r1 = psqrt( ((c_origin.vx*c_origin.vx)>>FIXED_PRECISION) +
			((c_origin.vy*c_origin.vy)>>FIXED_PRECISION) +
			((c_origin.vz*c_origin.vz)>>FIXED_PRECISION) );

		r2 = psqrt( ((c_origin2.vx*c_origin2.vx)>>FIXED_PRECISION) +
			((c_origin2.vy*c_origin2.vy)>>FIXED_PRECISION) +
			((c_origin2.vz*c_origin2.vz)>>FIXED_PRECISION) );

		FntPrint("\nr == %d\n", r);
		FntPrint("\n1 == %d\n", r1);
		FntPrint("\n2 == %d\n\n", r2);

		if(r < r1 && r > r2)
		{
			return 1;
		}
	}

	return 0;
}

/*
 * dgx_col_cxc:
 *	Does aabb collision detection/response of a CXC against:
 *		--another CXC
 *		--an interactive entity
 *
 *	Any active area triggers will activate a corresponding bit
 *	setting (via p_bitmask) relative to the initial CXC handler (p_cxc).
 */
dgx_col_cxc(p_cxc, p_bitmask, p_collider, collider_type)
struct AssetCXC *p_cxc;
int *p_bitmask;
void *p_collider;
int collider_type;
{
	int i, j;

	/* cxc-specific */
	unsigned short		*p_parent_scale;
	SVECTOR			*p_parent_global_pos;

	/* collider-specific */
	struct Entity3DI	*p_collider_e3di;
	struct AssetCXC		*p_collider_cxc;

	/* FIXME: streamline this dereferencing code instead of pasting it everywhere */
	switch(p_cxc->parent)
	{
		case CXC_PARENT_NONE:
		{
			struct AssetCXC *p = p_cxc;
			p_parent_scale = &p->ent.scale;
			p_parent_global_pos = &p->ent.global_pos;
			break;
		}
		case CXC_PARENT_MTD:
		{
			struct AssetMTD *p = (struct AssetMTD *)p_cxc->p_parent;
			p_parent_scale = &p->ent.scale;
			p_parent_global_pos = &p->ent.global_pos;
			break;
		}
		case CXC_PARENT_ATD:
		{
			struct AssetATD *p = (struct AssetATD *)p_cxc->p_parent;
			p_parent_scale = &p->ent.scale;
			p_parent_global_pos = &p->ent.global_pos;
			break;
		}
		default:
			printf("cxc_collision() error: unknown CXC parent (%d) !\n", p_cxc->parent);
			while(1);
	}

	/* dereference collider */
	switch(collider_type)
	{
		case CXC_COLLIDER_E3DI:
			p_collider_e3di = (struct Entity3DI *)p_collider;
			break;
		case CXC_COLLIDER_CXC:
			p_collider_cxc = (struct AssetCXC *)p_collider;
			break;
		default:
			printf("cxc_collision() error: unknown CXC collider (%d) !\n", collider_type);
			while(1);
	}

	/* COLMAP -> QUAD SLOPES */
	if(collider_type == CXC_COLLIDER_E3DI) /* slopes work exclusively with points */
	{
		for(i = 0; i < p_cxc->COLMAP.nquads; i++)
		{
			short		aabb_x[2];
			short		aabb_y[2];
			short		aabb_z[2];

			/* transform */
			for(j = 0; j < 2; j++)
			{
				aabb_x[j] = p_cxc->COLMAP.p_quad[i].aabb_x[j] * (*p_parent_scale);
				aabb_y[j] = p_cxc->COLMAP.p_quad[i].aabb_y[j] * (*p_parent_scale);
				aabb_z[j] = p_cxc->COLMAP.p_quad[i].aabb_z[j] * (*p_parent_scale);

				aabb_x[j] += p_parent_global_pos->vx;
				aabb_y[j] += p_parent_global_pos->vy;
				aabb_z[j] += p_parent_global_pos->vz;

				aabb_x[j] = -aabb_x[j];
				aabb_y[j] = -aabb_y[j];
				aabb_z[j] = -aabb_z[j];
			}

			/* is the player within the 2d bounding box (x & z)? */
			if( (p_collider_e3di->ent.global_pos.vx > aabb_x[0]) && (p_collider_e3di->ent.global_pos.vx < aabb_x[1]) )
			{
				if( (p_collider_e3di->ent.global_pos.vz < aabb_z[0]) && (p_collider_e3di->ent.global_pos.vz > aabb_z[1]) )
				{
					int t;

					switch( p_cxc->COLMAP.p_quad[i].global_coords.direction )
					{
						case CXC_SLOPE_FLAT:
							goto LERP_SKIP;
						case CXC_SLOPE_X:
						{
							int v1_n; /* normalized v1 */

							FntPrint("CXC_SLOPE_X (%d)\n", p_cxc->COLMAP.p_quad[i].global_coords.is_flipped);
							FntPrint("aabb_x == (%d, %d)\n", aabb_x[0], aabb_x[1]);
							FntPrint("aabb_y == (%d, %d)\n", aabb_y[0], aabb_y[1]);

							/*
							 * v0 == aabb_x[0]
							 * v1 == aabb_x[1]
							 */

							v1_n = aabb_x[0] - aabb_x[1];

							if(p_cxc->COLMAP.p_quad[i].global_coords.is_flipped > 0)
							{
								t = ((((int)p_collider_e3di->ent.global_pos.vx - aabb_x[1])<<FIXED_PRECISION) / v1_n);
							}
							else
							{
								t = (((aabb_x[0] - (int)p_collider_e3di->ent.global_pos.vx)<<FIXED_PRECISION) / v1_n);
							}

							if(!p_collider_e3di->is_jumping)
							{
								p_collider_e3di->is_on_slope = 1;
							}

							break;
						}
						case CXC_SLOPE_Z:
						{
							int v1_n; /* normalized v1 */

							FntPrint("CXC_SLOPE_Z (%d)\n", p_cxc->COLMAP.p_quad[i].global_coords.is_flipped);

							/*
							 * v0 == aabb_z[0]
							 * v1 == aabb_z[1]
							 */

							v1_n = aabb_z[0] - aabb_z[1];
							t = (((aabb_z[0] - (int)p_collider_e3di->ent.global_pos.vz)<<FIXED_PRECISION) / v1_n);

							if(!p_collider_e3di->is_jumping)
							{
								p_collider_e3di->is_on_slope = 1;
							}

							break;
						}
						default:
							printf("collision error: unknown slope direction on given quad!\n");
							printf("direction == %d\n", p_cxc->COLMAP.p_quad[i].global_coords.direction);
							printf("quad #%d\n", i);
							while(1);
					}

					p_collider_e3di->min_ground_pos = lerp(	((int)(aabb_y[0])),
										((int)(aabb_y[1])),
										t);

					p_collider_e3di->min_ground_pos >>= FIXED_PRECISION;

					LERP_SKIP:
				}
			}
		}
	}

	/* COLMAP -> CUBOID */
	for(i = 0; i < p_cxc->COLMAP.ncuboids; i++)
	{
		short		aabb_x[2];
		short		aabb_y[2];
		short		aabb_z[2];

		/* transform */
		for(j = 0; j < 2; j++)
		{
			aabb_x[j] = p_cxc->COLMAP.p_cuboid[i].aabb_x[j] * (*p_parent_scale);
			aabb_y[j] = p_cxc->COLMAP.p_cuboid[i].aabb_y[j] * (*p_parent_scale);
			aabb_z[j] = p_cxc->COLMAP.p_cuboid[i].aabb_z[j] * (*p_parent_scale);

			/* translate */
			aabb_x[j] += p_parent_global_pos->vx;
			aabb_y[j] += p_parent_global_pos->vy;
			aabb_z[j] += p_parent_global_pos->vz;

			aabb_x[j] = -aabb_x[j];
			aabb_y[j] = -aabb_y[j];
			aabb_z[j] = -aabb_z[j];
		}

		/*
		 * cuboid collision explanation:
		 *
		 *	First check whether the player is aligned to the 2D plane associated
		 *	with a particular axis. e.g. If doing z-axis collision, check the plane
		 *	located in the x and y axes.
		 *
		 *	If we're on the same side as that cuboid's plane AND if
		 *	we are to pass through the wall, alter velocity in that direction
		 *	to kiss the wall.
		 */

		{
			int is_cuboid_aligned_x = 0, is_cuboid_aligned_y = 0, is_cuboid_aligned_z = 0;

			switch(collider_type)
			{
				case CXC_COLLIDER_E3DI:
				{
					if( (p_collider_e3di->ent.global_pos.vx < aabb_x[0]) && (p_collider_e3di->ent.global_pos.vx > aabb_x[1]) )
					{
						is_cuboid_aligned_x = 1;
					}
					if( (p_collider_e3di->ent.global_pos.vy > aabb_y[0]) && (p_collider_e3di->ent.global_pos.vy < aabb_y[1]) )
					{
						is_cuboid_aligned_y = 1;
					}
					if( (p_collider_e3di->ent.global_pos.vz < aabb_z[0]) && (p_collider_e3di->ent.global_pos.vz > aabb_z[1]) )
					{
						is_cuboid_aligned_z = 1;
					}

					/* x-axis collision */
					if( is_cuboid_aligned_y )
					{
						if(is_cuboid_aligned_z)
						{
							if( p_collider_e3di->ent.global_pos.vx >= aabb_x[0] &&
								(p_collider_e3di->ent.global_pos.vx + p_collider_e3di->global_vel.vx < aabb_x[0]) )
							{
								p_collider_e3di->global_vel.vx = -p_collider_e3di->ent.global_pos.vx + aabb_x[0];
							}
							else if( p_collider_e3di->ent.global_pos.vx <= aabb_x[1] &&
								(p_collider_e3di->ent.global_pos.vx + p_collider_e3di->global_vel.vx > aabb_x[1]) )
							{
								p_collider_e3di->global_vel.vx = -p_collider_e3di->ent.global_pos.vx + aabb_x[1];
							}
						}
					}

					/* y-axis collision */
					if( is_cuboid_aligned_x )
					{
						if( is_cuboid_aligned_z )
						{
							if( p_collider_e3di->ent.global_pos.vy <= aabb_y[0] &&
								(p_collider_e3di->ent.global_pos.vy + p_collider_e3di->global_vel.vy > aabb_y[0]) )
							{
								p_collider_e3di->global_vel.vy = -p_collider_e3di->ent.global_pos.vy + aabb_y[0];
							}
							else if( p_collider_e3di->ent.global_pos.vy >= aabb_y[1] &&
								(p_collider_e3di->ent.global_pos.vy + p_collider_e3di->global_vel.vy < aabb_y[1]) )
							{
								p_collider_e3di->global_vel.vy = -p_collider_e3di->ent.global_pos.vy + aabb_y[1];
								p_collider_e3di->min_ground_pos = aabb_y[1];
							}
						}
					}

					/* z-axis collision */
					if( is_cuboid_aligned_x )
					{
						if( is_cuboid_aligned_y )
						{
							if( (p_collider_e3di->ent.global_pos.vz >= aabb_z[0]) && 
								(p_collider_e3di->ent.global_pos.vz + p_collider_e3di->global_vel.vz < aabb_z[0]) )
							{
								p_collider_e3di->global_vel.vz = -p_collider_e3di->ent.global_pos.vz + aabb_z[0];
							}
							else if( (p_collider_e3di->ent.global_pos.vz <= aabb_z[1]) && 
								(p_collider_e3di->ent.global_pos.vz + p_collider_e3di->global_vel.vz > aabb_z[1]) )
							{
								p_collider_e3di->global_vel.vz = -p_collider_e3di->ent.global_pos.vz + aabb_z[1];
							}
						}
					}
					break;
				}
			}
		}
	}

	/* AREATRIG -> CUBOID */
	for(i = 0; i < p_cxc->AREATRIG.ncuboids; i++)
	{
		short		aabb_x[2];
		short		aabb_y[2];
		short		aabb_z[2];

		short		collider_aabb_x[2];
		short		collider_aabb_y[2];
		short		collider_aabb_z[2];

		/* transform */
		for(j = 0; j < 2; j++)
		{
			aabb_x[j] = p_cxc->AREATRIG.p_cuboid[i].aabb_x[j] * (*p_parent_scale);
			aabb_y[j] = p_cxc->AREATRIG.p_cuboid[i].aabb_y[j] * (*p_parent_scale);
			aabb_z[j] = p_cxc->AREATRIG.p_cuboid[i].aabb_z[j] * (*p_parent_scale);

			/* translate */
			aabb_x[j] += p_parent_global_pos->vx;
			aabb_y[j] += p_parent_global_pos->vy;
			aabb_z[j] += p_parent_global_pos->vz;

			aabb_x[j] = -aabb_x[j];
			aabb_y[j] = -aabb_y[j];
			aabb_z[j] = -aabb_z[j];

			/* generate correct aabb's for cxc collider */
			if(collider_type == CXC_COLLIDER_CXC)
			{
				unsigned short	*p_scale;
				SVECTOR		*p_global_pos;

				/*
				 * IMPORTANT:
				 *	Only checks for 1 cuboid in collider CXC, since the collider CXC
				 *	does not necessarily have the same amount of cuboids as our main
				 *	CXC.
				 */
				collider_aabb_x[j] = p_collider_cxc->AREATRIG.p_cuboid[0].aabb_x[j];
				collider_aabb_y[j] = p_collider_cxc->AREATRIG.p_cuboid[0].aabb_y[j];
				collider_aabb_z[j] = p_collider_cxc->AREATRIG.p_cuboid[0].aabb_z[j];

				/* FIXME: streamline this dereferencing code instead of pasting it everywhere */
				switch(p_collider_cxc->parent)
				{
					case CXC_PARENT_NONE:
					{
						struct AssetCXC *p = p_collider_cxc;
						p_scale = &p->ent.scale;
						p_global_pos = &p->ent.global_pos;
						break;
					}
					case CXC_PARENT_MTD:
					{
						struct AssetMTD *p = (struct AssetMTD *)p_collider_cxc->p_parent;
						p_scale = &p->ent.scale;
						p_global_pos = &p->ent.global_pos;
						break;
					}
					case CXC_PARENT_ATD:
					{
						struct AssetATD *p = (struct AssetATD *)p_collider_cxc->p_parent;
						p_scale = &p->ent.scale;
						p_global_pos = &p->ent.global_pos;
						break;
					}
					default:
						printf("cxc_collision() error: unknown CXC parent (%d) !\n", p_collider_cxc->parent);
						while(1);
				}

				collider_aabb_x[j] *= *p_scale;
				collider_aabb_y[j] *= *p_scale;
				collider_aabb_z[j] *= *p_scale;

				collider_aabb_x[j] += p_global_pos->vx;
				collider_aabb_y[j] += p_global_pos->vy;
				collider_aabb_z[j] += p_global_pos->vz;

				collider_aabb_x[j] = -collider_aabb_x[j];
				collider_aabb_y[j] = -collider_aabb_y[j];
				collider_aabb_z[j] = -collider_aabb_z[j];
			}
		}

		/* are we inside the cuboid? */
		switch(collider_type)
		{
			case CXC_COLLIDER_E3DI:
			{
				if( (p_collider_e3di->ent.global_pos.vx < aabb_x[0]) && (p_collider_e3di->ent.global_pos.vx > aabb_x[1]) )
				{
					if( (p_collider_e3di->ent.global_pos.vy > aabb_y[0]) && (p_collider_e3di->ent.global_pos.vy < aabb_y[1]) )
					{
						if( (p_collider_e3di->ent.global_pos.vz < aabb_z[0]) && (p_collider_e3di->ent.global_pos.vz > aabb_z[1]) )
						{
							if(i < 32)
							{
								*p_bitmask |= (1<<i);
							}
							goto AREA_CUBOID_SKIP;
						}
					}
				}

				break;
			}
			case CXC_COLLIDER_CXC:
			{
				if( ((collider_aabb_x[0] <= aabb_x[0]) && (collider_aabb_x[0] >= aabb_x[1])) ||
					((collider_aabb_x[1] <= aabb_x[0]) && (collider_aabb_x[1] >= aabb_x[1])) )
				{
					if( ((collider_aabb_y[0] >= aabb_y[0]) && (collider_aabb_y[0] <= aabb_y[1])) ||
						((collider_aabb_y[1] >= aabb_y[0]) && (collider_aabb_y[1] <= aabb_y[1])) )
					{
						if( ((collider_aabb_z[0] <= aabb_z[0]) && (collider_aabb_z[0] >= aabb_z[1])) ||
							((collider_aabb_z[1] <= aabb_z[0]) && (collider_aabb_z[1] >= aabb_z[1])) )
						{
							if(i < 32)
							{
								*p_bitmask |= (1<<i);
							}
							goto AREA_CUBOID_SKIP;
						}
					}
				}

				break;
			}
		}

		/* we're not in cuboid, so remove it from bitmask */
		if( *p_bitmask & (1<<i) )
		{
			*p_bitmask -= (1<<i);
		}

		AREA_CUBOID_SKIP:
	}
}

